// This file is @generated by prost-build.
#[derive(Clone, PartialEq, ::prost::Message)]
pub struct Versions {
    /// All packages in the repository
    #[prost(message, repeated, tag = "1")]
    pub packages: ::prost::alloc::vec::Vec<VersionsPackage>,
    /// Name of repository
    #[prost(string, required, tag = "2")]
    pub repository: ::prost::alloc::string::String,
}
#[derive(Clone, PartialEq, ::prost::Message)]
pub struct VersionsPackage {
    /// Package name
    #[prost(string, required, tag = "1")]
    pub name: ::prost::alloc::string::String,
    /// All released versions of the package
    #[prost(string, repeated, tag = "2")]
    pub versions: ::prost::alloc::vec::Vec<::prost::alloc::string::String>,
    /// Zero-based indexes of retired versions in the versions field, see package.proto
    ///
    /// If set, the name of the package repository (NEVER USED, DEPRECATED)
    /// string repository = 4;
    #[prost(int32, repeated, tag = "3")]
    pub retired: ::prost::alloc::vec::Vec<i32>,
}
