// This file is @generated by prost-build.
#[derive(Clone, PartialEq, ::prost::Message)]
pub struct Signed {
    /// Signed contents
    #[prost(bytes = "vec", required, tag = "1")]
    pub payload: ::prost::alloc::vec::Vec<u8>,
    /// The signature
    #[prost(bytes = "vec", optional, tag = "2")]
    pub signature: ::core::option::Option<::prost::alloc::vec::Vec<u8>>,
}
