syntax = "proto2";

package names;

message Names {
  // All packages in the repository
  repeated Package packages = 1;
  // Name of repository
  required string repository = 2;
}

message Package {
  // Package name
  required string name = 1;

  // If set, the name of the package repository (NEVER USED, DEPRECATED)
  // string repository = 2;
}
